package de.westnordost.streetcomplete.screens.main.edithistory

import android.content.Context
import android.os.Bundle
import android.text.Html
import android.text.format.DateUtils
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.view.ViewGroup.LayoutParams.MATCH_PARENT
import android.view.ViewGroup.LayoutParams.WRAP_CONTENT
import android.widget.TextView
import androidx.appcompat.app.AlertDialog
import de.westnordost.osmfeatures.FeatureDictionary
import de.westnordost.streetcomplete.R
import de.westnordost.streetcomplete.data.edithistory.Edit
import de.westnordost.streetcomplete.data.edithistory.icon
import de.westnordost.streetcomplete.data.edithistory.overlayIcon
import de.westnordost.streetcomplete.data.osm.edits.ElementEdit
import de.westnordost.streetcomplete.data.osm.edits.create.CreateNodeAction
import de.westnordost.streetcomplete.data.osm.edits.create.CreateNodeFromVertexAction
import de.westnordost.streetcomplete.data.osm.edits.delete.DeletePoiNodeAction
import de.westnordost.streetcomplete.data.osm.edits.move.MoveNodeAction
import de.westnordost.streetcomplete.data.osm.edits.split_way.SplitWayAction
import de.westnordost.streetcomplete.data.osm.edits.update_tags.StringMapEntryAdd
import de.westnordost.streetcomplete.data.osm.edits.update_tags.StringMapEntryChange
import de.westnordost.streetcomplete.data.osm.edits.update_tags.StringMapEntryDelete
import de.westnordost.streetcomplete.data.osm.edits.update_tags.StringMapEntryModify
import de.westnordost.streetcomplete.data.osm.edits.update_tags.UpdateElementTagsAction
import de.westnordost.streetcomplete.data.osm.mapdata.Element
import de.westnordost.streetcomplete.data.osm.osmquests.OsmQuestHidden
import de.westnordost.streetcomplete.data.osmnotes.edits.NoteEdit
import de.westnordost.streetcomplete.data.osmnotes.edits.NoteEditAction.CLOSE
import de.westnordost.streetcomplete.data.osmnotes.edits.NoteEditAction.COMMENT
import de.westnordost.streetcomplete.data.osmnotes.edits.NoteEditAction.CREATE
import de.westnordost.streetcomplete.data.osmnotes.notequests.OsmNoteQuestHidden
import de.westnordost.streetcomplete.data.externalsource.ExternalSourceQuestHidden
import de.westnordost.streetcomplete.data.osm.edits.create.CreateRelationAction
import de.westnordost.streetcomplete.data.osm.edits.delete.DeleteRelationAction
import de.westnordost.streetcomplete.data.quest.QuestType
import de.westnordost.streetcomplete.databinding.DialogUndoBinding
import de.westnordost.streetcomplete.quests.getHtmlQuestTitle
import de.westnordost.streetcomplete.util.getNameAndLocationLabel
import de.westnordost.streetcomplete.util.ktx.nowAsEpochMilliseconds
import de.westnordost.streetcomplete.view.CharSequenceText
import de.westnordost.streetcomplete.view.ResText
import de.westnordost.streetcomplete.view.Text
import de.westnordost.streetcomplete.view.setHtml
import de.westnordost.streetcomplete.view.setText
import java.util.MissingFormatArgumentException

class UndoDialog(
    context: Context,
    private val edit: Edit,
    private val element: Element?,
    private val featureDictionaryLazy: Lazy<FeatureDictionary>,
    private val onUndo: (edit: Edit) -> Unit,
) : AlertDialog(context) {

    private val binding = DialogUndoBinding.inflate(LayoutInflater.from(context))

    init {
        binding.icon.setImageResource(edit.icon)
        val overlayResId = edit.overlayIcon
        if (overlayResId != 0) binding.overlayIcon.setImageResource(overlayResId)
        binding.createdTimeText.text =
            DateUtils.getRelativeTimeSpanString(edit.createdTimestamp, nowAsEpochMilliseconds(), DateUtils.MINUTE_IN_MILLIS)
        binding.descriptionContainer.addView(edit.descriptionView)

        setTitle(R.string.undo_confirm_title2)
        setView(binding.root)
        setButton(BUTTON_POSITIVE, context.getText(R.string.undo_confirm_positive), null) { _, _ -> onUndo(edit) }
        setButton(BUTTON_NEGATIVE, context.getText(R.string.undo_confirm_negative), null, null)
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        binding.titleText.text = edit.getTitle()
        if (edit is ElementEdit) {
            binding.titleHintText.text = element?.let {
                getNameAndLocationLabel(it, context.resources, featureDictionaryLazy.value)
            }
        }
    }

    private fun Edit.getTitle(): CharSequence = when (this) {
        is ElementEdit -> {
            if (type is QuestType) {
                getQuestTitle(type, element?.tags.orEmpty())
            } else {
                context.resources.getText(type.title)
            }
        }
        is NoteEdit -> {
            context.resources.getText(when (action) {
                CREATE -> R.string.created_note_action_title
                COMMENT -> R.string.commented_note_action_title
                CLOSE -> R.string.closed_note_action_title
            })
        }
        is OsmQuestHidden -> {
            getQuestTitle(questType, element?.tags.orEmpty())
        }
        is OsmNoteQuestHidden -> {
            context.resources.getText(R.string.quest_noteDiscussion_title)
        }
        is ExternalSourceQuestHidden -> {
            getQuestTitle(questType, emptyMap())
        }
        else -> throw IllegalArgumentException()
    }

    private val Edit.descriptionView: View get() = when (this) {
        is ElementEdit -> {
            when (action) {
                is UpdateElementTagsAction ->
                    createListOfTagUpdates(action.changes.changes)
                is DeletePoiNodeAction, is DeleteRelationAction ->
                    createTextView(ResText(R.string.deleted_poi_action_description))
                is SplitWayAction ->
                    createTextView(ResText(R.string.split_way_action_description))
                is CreateNodeAction ->
                    createCreateNodeDescriptionView(action.tags)
                is CreateNodeFromVertexAction ->
                    createListOfTagUpdates(action.changes.changes)
                is MoveNodeAction ->
                    createTextView(ResText(R.string.move_node_action_description))
                is CreateRelationAction ->
                    createCreateRelationDescriptionView(action.tags)
                else -> throw IllegalArgumentException()
            }
        }
        is NoteEdit -> createTextView(text?.let { CharSequenceText(it) })
        is OsmQuestHidden -> createTextView(ResText(R.string.hid_action_description))
        is OsmNoteQuestHidden, is ExternalSourceQuestHidden -> createTextView(ResText(R.string.hid_action_description))
        else -> throw IllegalArgumentException()
    }

    private fun getQuestTitle(questType: QuestType, tags: Map<String, String>): CharSequence =
        try {
            context.resources.getHtmlQuestTitle(questType, tags)
        } catch (e: MissingFormatArgumentException) {
            /* The exception happens when the number of format strings in the quest title
             * differs from what can be "filled" by getHtmlQuestTitle. When does this happen?
             * It happens the element is null or otherwise is not at all what is expected by
             * that quest type.
             * So, this is the fallback for that case */
            context.resources.getString(questType.title, *Array(10) { "…" })
        }

    private fun createTextView(text: Text?): TextView {
        val txt = TextView(context)
        txt.layoutParams = ViewGroup.LayoutParams(MATCH_PARENT, WRAP_CONTENT)
        txt.setText(text)
        txt.setTextIsSelectable(true)
        return txt
    }

    private fun createListOfTagUpdates(changes: Collection<StringMapEntryChange>): TextView {
        val txt = TextView(context)
        txt.layoutParams = ViewGroup.LayoutParams(MATCH_PARENT, WRAP_CONTENT)

        txt.setHtml(changes.joinToString(separator = "", prefix = "<ul>", postfix = "</ul>") { change ->
           "<li>" +
           context.resources.getString(
               change.titleResId,
               "<tt>" + change.toLinkedTagString() + "</tt>"
           ) +
           "</li>"
        })
        return txt
    }

    private fun createCreateNodeDescriptionView(tags: Map<String, String>): TextView {
        val txt = TextView(context)
        txt.layoutParams = ViewGroup.LayoutParams(MATCH_PARENT, WRAP_CONTENT)

        txt.setHtml(
            context.resources.getString(R.string.create_node_action_description) +
            tags.entries.joinToString(separator = "", prefix = "<ul>", postfix = "</ul>") { (key, value) ->
                "<li><tt>" + linkedTagString(key, value) + "</tt></li>"
            }
        )
        return txt
    }
    private fun createCreateRelationDescriptionView(tags: Map<String, String>): TextView {
        val txt = TextView(context)
        txt.layoutParams = ViewGroup.LayoutParams(MATCH_PARENT, WRAP_CONTENT)

        txt.setHtml(
        context.resources.getString(R.string.create_relation_action_description) +
            tags.entries.joinToString(separator = "", prefix = "<ul>", postfix = "</ul>") { (key, value) ->
                "<li><tt>" + linkedTagString(key, value) + "</tt></li>"
            }
        )
        return txt
    }
}

private fun StringMapEntryChange.toLinkedTagString(): String =
    linkedTagString(key, when (this) {
        is StringMapEntryAdd -> value
        is StringMapEntryModify -> value
        is StringMapEntryDelete -> valueBefore
    })

private fun linkedTagString(key: String, value: String): String {
    val escapedKey = Html.escapeHtml(key)
    val escapedValue = Html.escapeHtml(value)
    val keyLink = "<a href=\"https://wiki.openstreetmap.org/wiki/Key:$escapedKey\">$escapedKey</a>"
    return "$keyLink = $escapedValue"
}

private val StringMapEntryChange.titleResId: Int get() = when (this) {
    is StringMapEntryAdd -> R.string.added_tag_action_title
    is StringMapEntryModify -> R.string.changed_tag_action_title
    is StringMapEntryDelete -> R.string.removed_tag_action_title
}
