package de.westnordost.streetcomplete.screens.main.map

import android.content.Context
import android.graphics.Bitmap
import android.graphics.Canvas
import com.russhwolf.settings.ObservableSettings
import de.westnordost.streetcomplete.Prefs
import de.westnordost.streetcomplete.R
import de.westnordost.streetcomplete.data.overlays.OverlayRegistry
import de.westnordost.streetcomplete.data.quest.QuestTypeRegistry
import de.westnordost.streetcomplete.util.ktx.isApril1st
import kotlin.math.ceil
import kotlin.math.sqrt

/** From all the quest types, creates and saves a sprite sheet of quest type pin icons, provides
 *  the scene updates for tangram to access this sprite sheet  */
class TangramPinsSpriteSheet(
    private val context: Context,
    private val questTypeRegistry: QuestTypeRegistry,
    private val overlayRegistry: OverlayRegistry,
    private val prefs: ObservableSettings
) {
    val sceneUpdates: List<Pair<String, String>> by lazy {
        val lastUpdate = context.packageManager.getPackageInfo(context.packageName, 0).lastUpdateTime.toInt()
        val isSpriteSheetCurrent = prefs.getInt(Prefs.PIN_SPRITES_VERSION, 0) == lastUpdate

        val spriteSheet = when {
            !isSpriteSheetCurrent || shouldBeUpsideDown() -> createSpritesheet()
            else -> prefs.getStringOrNull(Prefs.PIN_SPRITES) ?: ""
        }

        createSceneUpdates(spriteSheet)
    }

    private fun createSpritesheet(): String {
        val questIconResIds = (
            questTypeRegistry.map { it.icon } +
            overlayRegistry.map { it.icon } +
            ADDITIONAL_ICONS
        ).toSortedSet()

        val spriteSheetEntries: MutableList<String> = ArrayList(questIconResIds.size)
        val questPin = context.resources.getDrawable(R.drawable.pin)
        val iconSize = questPin.intrinsicWidth
        val questIconSize = 2 * iconSize / 3
        val questIconOffsetX = 56 * iconSize / 192
        val questIconOffsetY = 18 * iconSize / 192
        val sheetSideLength = ceil(sqrt(questIconResIds.size.toDouble())).toInt()
        val bitmapLength = sheetSideLength * iconSize
        val spriteSheet = Bitmap.createBitmap(bitmapLength, bitmapLength, Bitmap.Config.ARGB_8888)
        val canvas = Canvas(spriteSheet)

        for ((i, questIconResId) in questIconResIds.withIndex()) {
            val x = i % sheetSideLength * iconSize
            val y = i / sheetSideLength * iconSize
            questPin.setBounds(x, y, x + iconSize, y + iconSize)
            questPin.draw(canvas)
            val questIcon = context.getDrawable(questIconResId)!!
            val questX = x + questIconOffsetX
            val questY = y + questIconOffsetY
            questIcon.setBounds(questX, questY, questX + questIconSize, questY + questIconSize)
            val checkpoint = canvas.save()
            if (shouldBeUpsideDown()) {
                val questCenterX = questX + questIconSize / 2f
                val questCenterY = questY + questIconSize / 2f
                canvas.rotate(180f, questCenterX, questCenterY)
            }
            questIcon.draw(canvas)
            canvas.restoreToCount(checkpoint)
            val questIconName = context.resources.getResourceEntryName(questIconResId)
            spriteSheetEntries.add("$questIconName: [$x,$y,$iconSize,$iconSize]")
        }

        context.deleteFile(PIN_ICONS_FILE)
        val spriteSheetIconsFile = context.openFileOutput(PIN_ICONS_FILE, Context.MODE_PRIVATE)
        spriteSheet.compress(Bitmap.CompressFormat.PNG, 0, spriteSheetIconsFile)
        spriteSheetIconsFile.close()

        val questSprites = "{${spriteSheetEntries.joinToString(",")}}"

        prefs.putInt(Prefs.PIN_SPRITES_VERSION, if (shouldBeUpsideDown()) -1 else context.packageManager.getPackageInfo(context.packageName, 0).lastUpdateTime.toInt())
        prefs.putString(Prefs.PIN_SPRITES, questSprites)

        return questSprites
    }

    private fun shouldBeUpsideDown(): Boolean {
        val isBelowEquator = prefs.getDouble(Prefs.MAP_LATITUDE, 0.0) < 0.0
        return isBelowEquator && isApril1st()
    }

    private fun createSceneUpdates(pinSprites: String): List<Pair<String, String>> = listOf(
        "textures.pins.url" to "file://${context.filesDir}/$PIN_ICONS_FILE",
        "textures.pins.sprites" to pinSprites
    )

    companion object {
        private const val PIN_ICONS_FILE = "pins.png"
        private val ADDITIONAL_ICONS = listOf(
            R.drawable.ic_quest_create_note,
            R.drawable.ic_quest_close_note,
            R.drawable.ic_add_poi,
            R.drawable.ic_edit_tags
        )
    }
}
